"""A test JSON parser"""

import json
from pegasus import Parser, rule
from pegasus.rules import *
from pegasus.rules import ChrRange as C
from pegasus.rules import Discard as _
from pegasus.util import flatten


ESCAPES = {
    'r': '\r',
    'n': '\n',
    'v': '\v',
    't': '\t',
    'b': '\b',
    'a': '\a',
    'f': '\f',
    '0': '\0',
    '"': '"',
    '\\': '\\'
}


class JsonParser(Parser):
    @rule(Star(In(' \t\r\n\f')))
    def ws(self, *_):
        pass

    @rule('null')
    def null_literal(self, *_):
        return (None,)

    @rule('true')
    def true_literal(self, *_):
        return True

    @rule('false')
    def false_literal(self, *_):
        return False

    @rule([true_literal, false_literal])
    def bool_literal(self, boolean):
        return boolean

    @rule(Plus(C['0':'9']))
    def digits(self, *digits):
        return digits

    @rule(Str(Opt(['+', '-']), [(digits, Opt('.', Opt(digits))), (Opt(digits), '.', digits)]))
    def number(self, number):
        return float(number)

    @rule([C['0':'9'], C['A':'F'], C['a':'f']])
    def hex_char(self, char):
        return char

    @rule(Str(hex_char, hex_char))
    def hex8(self, chars):
        return int(chars, 16)

    @rule(Str(hex_char, hex_char, hex_char, hex_char))
    def hex16(self, chars):
        return int(chars, 16)

    @rule(_('\\'), [In(ESCAPES.keys()), ('x', hex8), ('u', hex16)])
    def char_escape(self, seq, num=0):
        if seq in 'xu':
            return chr(num)

        return ESCAPES[seq]

    @rule(_('"'), Str(Star([char_escape, In('\\"', True)])), _('"'))
    def string(self, contents):
        return contents

    @rule([string, number, bool_literal, null_literal, Lazy('array'), Lazy('object')])
    def value(self, value):
        return value

    @rule(_('[', ws), Opt(value, Star(_(ws, ',', ws), value)), _(ws, ']'))
    def array(self, first=None, *rest):
        if first is None and (rest is None or not len(rest)):
            return []
        return list((first,) + tuple(flatten(rest, depth=1)))

    @rule(string, _(ws, ':', ws), value)
    def object_pair(self, key, val):
        return (key, val)

    @rule(_('{', ws), Opt(object_pair, Star(_(ws, ',', ws), object_pair)), _(ws, '}'))
    def object(self, first=None, *rest):
        if first is None and (rest is None or not len(rest)):
            return {}

        # ugh, don't we just love Python? I mean seriously... <3
        return dict((first,) + tuple(flatten(rest, depth=1)))

    @rule(ws, value, ws, EOF)
    def document(self, document):
        return document


def test_json_boolean():
    parser = JsonParser()
    assert parser.parse(JsonParser.true_literal, 'true', match=False) is True
    assert parser.parse(JsonParser.false_literal, 'false', match=False) is False
    assert parser.parse(JsonParser.bool_literal, 'true', match=False) is True
    assert parser.parse(JsonParser.bool_literal, 'false', match=False) is False


def test_json_number():
    parser = JsonParser()
    assert parser.parse(JsonParser.number, '1234', match=False) == 1234.0
    assert parser.parse(JsonParser.number, '1234.', match=False) == 1234.0
    assert parser.parse(JsonParser.number, '1234.5678', match=False) == 1234.5678
    assert parser.parse(JsonParser.number, '.1234', match=False) == 0.1234
    assert parser.parse(JsonParser.number, '+1234', match=False) == 1234.0
    assert parser.parse(JsonParser.number, '+1234.', match=False) == 1234.0
    assert parser.parse(JsonParser.number, '+.1234', match=False) == 0.1234
    assert parser.parse(JsonParser.number, '+1234.5678', match=False) == 1234.5678
    assert parser.parse(JsonParser.number, '-1234', match=False) == -1234.0
    assert parser.parse(JsonParser.number, '-1234.', match=False) == -1234.0
    assert parser.parse(JsonParser.number, '-.1234', match=False) == -0.1234
    assert parser.parse(JsonParser.number, '-1234.5678', match=False) == -1234.5678


def test_json_string():
    parser = JsonParser()
    assert parser.parse(JsonParser.string, '"hello"', match=False) == 'hello'
    assert parser.parse(JsonParser.string, '"hello there"', match=False) == 'hello there'
    assert parser.parse(JsonParser.string, '"\n"', match=False) == '\n'
    assert parser.parse(JsonParser.string, '"\\\\"', match=False) == '\\'
    assert parser.parse(JsonParser.string, '"\\\\\\""', match=False) == '\\"'
    assert parser.parse(JsonParser.string, '"\\v\\t\\n"', match=False) == '\v\t\n'


def test_json_null():
    parser = JsonParser()
    assert parser.parse(JsonParser.null_literal, 'null', match=False) == (None,)


def test_json_array():
    parser = JsonParser()
    assert parser.parse(JsonParser.array, '[1]', match=False) == [1]
    assert parser.parse(JsonParser.array, '[1, 2, 3]', match=False) == [1, 2, 3]
    assert parser.parse(JsonParser.array, '[1,2, \n3]', match=False) == [1, 2, 3]
    assert parser.parse(JsonParser.array, '[1, 2, \n3, true, false]', match=False) == [1, 2, 3, True, False]
    assert parser.parse(JsonParser.array, '[]', match=False) == []


def test_json_nested_array():
    parser = JsonParser()
    assert parser.parse(JsonParser.array, '[[]]', match=False) == [[]]
    assert parser.parse(JsonParser.array, '[[[]]]', match=False) == [[[]]]
    assert parser.parse(JsonParser.array, '[[[],[]]]', match=False) == [[[], []]]
    assert parser.parse(JsonParser.array, '[[1]]', match=False) == [[1]]
    assert parser.parse(JsonParser.array, '[[1],[2, 3]]', match=False) == [[1], [2, 3]]


def test_json_object():
    parser = JsonParser()
    assert parser.parse(JsonParser.object, '{}', match=False) == {}
    assert parser.parse(JsonParser.object, '{"foo": true}', match=False) == {'foo': True}
    assert parser.parse(JsonParser.object, '{"foo": true, "hello": 12345}', match=False) == {'foo': True, 'hello': 12345.0}
    assert parser.parse(JsonParser.object, '{"foo": true, "hello": 12345, "another": [1, 2, 3]}', match=False) == {'foo': True, 'hello': 12345.0, 'another': [1, 2, 3]}


def test_json_value():
    parser = JsonParser()
    assert parser.parse(JsonParser.value, 'null', match=False) == (None,)
    assert parser.parse(JsonParser.value, '"hello, there!"', match=False) == 'hello, there!'
    assert parser.parse(JsonParser.value, '1234.5678', match=False) == 1234.5678
    assert parser.parse(JsonParser.value, '.5678', match=False) == 0.5678
    assert parser.parse(JsonParser.value, '1234.', match=False) == 1234.0
    assert parser.parse(JsonParser.value, '1234', match=False) == 1234.0
    assert parser.parse(JsonParser.value, 'true', match=False) is True
    assert parser.parse(JsonParser.value, 'false', match=False) is False
    assert parser.parse(JsonParser.value, '[12345]', match=False) == [12345.0]
    assert parser.parse(JsonParser.value, '[[12345], true]', match=False) == [[12345.0], True]


def test_json_documents():
    parser = JsonParser()

    def validate(doc):
        return json.loads(doc) == parser.parse(JsonParser.document, doc)

    validate("""
        {
            "glossary": {
                "title": "example glossary",
                "GlossDiv": {
                    "title": "S",
                    "GlossList": {
                        "GlossEntry": {
                            "ID": "SGML",
                            "SortAs": "SGML",
                            "GlossTerm": "Standard Generalized Markup Language",
                            "Acronym": "SGML",
                            "Abbrev": "ISO 8879:1986",
                            "GlossDef": {
                                "para": "A meta-markup language, used to create markup languages such as DocBook.",
                                "GlossSeeAlso": ["GML", "XML"]
                            },
                            "GlossSee": "markup"
                        }
                    }
                }
            }
        }
    """)

    validate("""
        {"menu": {
          "id": "file",
          "value": "File",
          "popup": {
            "menuitem": [
              {"value": "New", "onclick": "CreateNewDoc()"},
              {"value": "Open", "onclick": "OpenDoc()"},
              {"value": "Close", "onclick": "CloseDoc()"}
            ]
          }
        }}
    """)

    validate("""
            {"widget": {
            "debug": "on",
            "window": {
                "title": "Sample Konfabulator Widget",
                "name": "main_window",
                "width": 500,
       	"height": 500
            },
            "image": {
                "src": "Images/Sun.png",
                "name": "sun1",
                "hOffset": 250,
                "vOffset": 250,
                "alignment": "center"
            },
            "text": {
                "data": "Click Here",
                "size": 36,
                "style": "bold",
                "name": "text1",
                "hOffset": 250,
                "vOffset": 100,
                "alignment": "center",
                "onMouseUp": "sun1.opacity = (sun1.opacity / 100) * 90;"
            }
        }}
    """)

    validate("""
{"web-app": {
  "servlet": [
    {
      "servlet-name": "cofaxCDS",
      "servlet-class": "org.cofax.cds.CDSServlet",
      "init-param": {
        "configGlossary:installationAt": "Philadelphia, PA",
        "configGlossary:adminEmail": "ksm@pobox.com",
        "configGlossary:poweredBy": "Cofax",
        "configGlossary:poweredByIcon": "/images/cofax.gif",
        "configGlossary:staticPath": "/content/static",
        "templateProcessorClass": "org.cofax.WysiwygTemplate",
        "templateLoaderClass": "org.cofax.FilesTemplateLoader",
        "templatePath": "templates",
        "templateOverridePath": "",
        "defaultListTemplate": "listTemplate.htm",
        "defaultFileTemplate": "articleTemplate.htm",
        "useJSP": false,
        "jspListTemplate": "listTemplate.jsp",
        "jspFileTemplate": "articleTemplate.jsp",
        "cachePackageTagsTrack": 200,
        "cachePackageTagsStore": 200,
        "cachePackageTagsRefresh": 60,
        "cacheTemplatesTrack": 100,
        "cacheTemplatesStore": 50,
        "cacheTemplatesRefresh": 15,
        "cachePagesTrack": 200,
        "cachePagesStore": 100,
        "cachePagesRefresh": 10,
        "cachePagesDirtyRead": 10,
        "searchEngineListTemplate": "forSearchEnginesList.htm",
        "searchEngineFileTemplate": "forSearchEngines.htm",
        "searchEngineRobotsDb": "WEB-INF/robots.db",
        "useDataStore": true,
        "dataStoreClass": "org.cofax.SqlDataStore",
        "redirectionClass": "org.cofax.SqlRedirection",
        "dataStoreName": "cofax",
        "dataStoreDriver": "com.microsoft.jdbc.sqlserver.SQLServerDriver",
        "dataStoreUrl": "jdbc:microsoft:sqlserver://LOCALHOST:1433;DatabaseName=goon",
        "dataStoreUser": "sa",
        "dataStorePassword": "dataStoreTestQuery",
        "dataStoreTestQuery": "SET NOCOUNT ON;select test='test';",
        "dataStoreLogFile": "/usr/local/tomcat/logs/datastore.log",
        "dataStoreInitConns": 10,
        "dataStoreMaxConns": 100,
        "dataStoreConnUsageLimit": 100,
        "dataStoreLogLevel": "debug",
        "maxUrlLength": 500}},
    {
      "servlet-name": "cofaxEmail",
      "servlet-class": "org.cofax.cds.EmailServlet",
      "init-param": {
      "mailHost": "mail1",
      "mailHostOverride": "mail2"}},
    {
      "servlet-name": "cofaxAdmin",
      "servlet-class": "org.cofax.cds.AdminServlet"},

    {
      "servlet-name": "fileServlet",
      "servlet-class": "org.cofax.cds.FileServlet"},
    {
      "servlet-name": "cofaxTools",
      "servlet-class": "org.cofax.cms.CofaxToolsServlet",
      "init-param": {
        "templatePath": "toolstemplates/",
        "log": 1,
        "logLocation": "/usr/local/tomcat/logs/CofaxTools.log",
        "logMaxSize": "",
        "dataLog": 1,
        "dataLogLocation": "/usr/local/tomcat/logs/dataLog.log",
        "dataLogMaxSize": "",
        "removePageCache": "/content/admin/remove?cache=pages&id=",
        "removeTemplateCache": "/content/admin/remove?cache=templates&id=",
        "fileTransferFolder": "/usr/local/tomcat/webapps/content/fileTransferFolder",
        "lookInContext": 1,
        "adminGroupID": 4,
        "betaServer": true}}],
  "servlet-mapping": {
    "cofaxCDS": "/",
    "cofaxEmail": "/cofaxutil/aemail/*",
    "cofaxAdmin": "/admin/*",
    "fileServlet": "/static/*",
    "cofaxTools": "/tools/*"},

  "taglib": {
    "taglib-uri": "cofax.tld",
    "taglib-location": "/WEB-INF/tlds/cofax.tld"}}}
    """)

    validate("""
{"menu": {
    "header": "SVG Viewer",
    "items": [
        {"id": "Open"},
        {"id": "OpenNew", "label": "Open New"},
        null,
        {"id": "ZoomIn", "label": "Zoom In"},
        {"id": "ZoomOut", "label": "Zoom Out"},
        {"id": "OriginalView", "label": "Original View"},
        null,
        {"id": "Quality"},
        {"id": "Pause"},
        {"id": "Mute"},
        null,
        {"id": "Find", "label": "Find..."},
        {"id": "FindAgain", "label": "Find Again"},
        {"id": "Copy"},
        {"id": "CopyAgain", "label": "Copy Again"},
        {"id": "CopySVG", "label": "Copy SVG"},
        {"id": "ViewSVG", "label": "View SVG"},
        {"id": "ViewSource", "label": "View Source"},
        {"id": "SaveAs", "label": "Save As"},
        null,
        {"id": "Help"},
        {"id": "About", "label": "About Adobe CVG Viewer..."}
    ]
}}
    """)



"""Pegasus parser rules, in all their glory

Here are how rules work:

    All rules are generators. If a rule by itself isn't a generator function,
    it's intended to be called with some configuration parameters, and thus
    returns a generator function (e.g. Seq and Or). Otherwise, it is to be called
    directly in order to get a parser generator (e.g. EOF).

    The generator returned should yield a two-element tuple, the first element being one of two things:
        - None in the event the rule is still consuming but hasn't succeeded
        - (result,) in the event the rule has succeeded and has garnered a result.
          In the event the rule succeeds but no result is to be returned (e.g. the rule
          is marked as discarded), then an empty tuple should be returned.

    In the event the rule fails, a ParseError should be raised.

    In the event the rule consumed a character it shouldn't have, the second tuple element should be True.

    Lastly, in the event the rule succeeds without question (e.g. a successful string literal), simply a single
    element tuple with the result or a tuple of `(<result>, False)` should be returned.

    In the event the rule expires (a StopIteration is thrown), the exception should be propagated
    upward in order to show the full stack.

    When creating the rule generators, a `char` value is passed. This value is a callable that takes
    no arguments and returns the current character. It is NOT a generator nor does it modify any value,
    so it can be safely called multiple times to return the same character. However, once the function
    yields, a subsequent call to char() upon reentry will NOT return the same character!

    It is expected that upon a rule generator being initialized that it should read the current character.
    This is why Seq() iterates through rules via a generator rather than a list.
"""
import inspect
from pegasus.util import flatten


DEBUG = False
__dbgdepth = 0


def set_debug(debug=True):
    global DEBUG
    DEBUG = debug


def debuggable(name=None):
    def _wrap(fn):
        if not DEBUG:
            return fn

        _name = name if name else fn.__name__

        def _inner(char, *args, **kwargs):
            global __dbgdepth

            gen = fn(char, *args, **kwargs)

            depth = ' ' * __dbgdepth
            ch = repr(char())
            while True:
                try:
                    print 'pegasus: {}\x1b[2;38;5;241menter {} -> {}\x1b[m'.format(depth, ch, _name)
                    __dbgdepth += 1
                    result, reconsume = next(gen)
                    if result is not None:
                        print 'pegasus: {}\x1b[1;38;5;126mresult {} -> {} ==> {} (reconsume={})\x1b[m'.format(depth, ch, _name, result, reconsume)
                    yield result, reconsume
                except Exception as e:
                    print 'pegasus: {}\x1b[38;5;88mfail {} -> {}\t{}\x1b[m'.format(depth, ch, _name, str(e))
                    raise
                finally:
                    print 'pegasus: {}\x1b[2;38;5;241mexit {} -> {}\x1b[m'.format(depth, ch, _name)
                    __dbgdepth -= 1

        return _inner
    return _wrap


class BadRuleException(Exception):
    """Thrown if a rule was invalid, due to a bad type usually"""
    pass


class ParseError(Exception):
    """Thrown in the event there was a problem parsing the input string"""
    def __init__(self, got=None, expected=None):
        self.got = got
        self.expected = expected if expected else []

        rgot = repr(got)
        if got is not None and expected is None or len(expected) == 0:
            message = 'unexpected: {}'.format(rgot)
        elif got is None and expected is not None and len(expected):
            if len(expected) == 1:
                message = 'expected: {}'.format(expected[0])
            else:
                message = 'expected one of the following:'
                for e in expected:
                    message += '\n- {}'.format(e)
        elif got is None and expected is None:
            message = 'unknown parse error'
        else:
            if len(expected) == 1:
                message = 'got: {}, expected: {}'.format(rgot, expected[0])
            else:
                message = 'got: {}, expected one of:'.format(rgot)
                for e in expected:
                    message += '\n- {}'.format(e)

        super(ParseError, self).__init__(message)

    @classmethod
    def combine(cls, errors):
        expected = []
        for error in errors:
            if not len(error.expected):
                continue
            for exp in error.expected:
                expected.append('{} but got \'{}\' instead'.format(exp, error.got))

        return ParseError(expected=expected)


class Lazy(object):
    _LOOKUPS = {}

    def __init__(self, name):
        self.name = name
        stack = inspect.stack()[1]
        self.module = stack[3]
        self.file = stack[1]

    def resolve(self):
        if (self.file, self.module, self.name) not in Lazy._LOOKUPS:
            raise BadRuleException('could not resolve lazily loaded rule: {}.{}'.format(self.module, self.name))

        return Lazy._LOOKUPS[(self.file, self.module, self.name)]


def _build_rule(rule):
    if isinstance(rule, Lazy):
        rule = rule.resolve()

    if callable(rule):
        if hasattr(rule, '_rule'):
            # it's a class rule that has a transformation step
            return ParserRule(rule, getattr(rule, '_rule'))
        return rule

    if type(rule) in [str, unicode]:
        return Literal(rule)

    if type(rule) == list:
        if len(rule) == 0:
            raise BadRuleException('Or() rules must have at least one condition')
        return _build_rule(rule[0]) if len(rule) == 1 else Or(*rule)

    if type(rule) == tuple:
        if len(rule) == 0:
            raise BadRuleException('Seq() rules must have at least one condition')
        return _build_rule(rule[0]) if len(rule) == 1 else Seq(*rule)

    raise BadRuleException('rule has invalid type: {}'.format(repr(rule)))


@debuggable('EOF')
def EOF(char, parser):
    """Fails if the given character is not None"""
    if char() is not None:
        raise ParseError(got=char(), expected=['<EOF>'])

    yield (), False


def ParserRule(class_rule, parse_rule):
    """Calls a transformation step class_rule if the parse_rule succeeds"""
    rule = _build_rule(parse_rule)

    @debuggable('ParserRule')
    def _iter(char, parser):
        grule = rule(char, parser)

        while True:
            result, reconsume = next(grule)
            if result is not None:
                result = class_rule(parser, *result)
                yield ((result,) if result is not None else ()), reconsume
                break

            yield None, reconsume

    return _iter


def Literal(utf):
    """Matches an exact string literal"""
    if type(utf) == str:
        utf = unicode(utf)

    length = len(utf)

    @debuggable('Literal')
    def _iter(char, parser):
        for i in xrange(length):
            c = utf[i]
            if char() and c == char():
                if i + 1 == length:
                    break
                yield None, None
            else:
                raise ParseError(got=char() or '<EOF>', expected=['{} (in literal {})'.format(repr(c), repr(utf))])

        yield (utf,), False

    return _iter


def Or(*rules):
    """Matches the first succeeding rule"""

    @debuggable('Or')
    def _iter(char, parser):
        remaining = [_build_rule(rule)(char, parser) for rule in rules]
        errors = []

        while len(remaining):
            for rule in list(remaining):
                reconsume = True
                while reconsume:
                    try:
                        result, reconsume = next(rule)
                        if result is not None:
                            yield result, reconsume
                            raise StopIteration()
                    except ParseError as e:
                        errors.append(e)
                        remaining.remove(rule)
                        break

            if len(remaining):
                yield None, False

        raise ParseError.combine(errors)

    return _iter


def Seq(*rules):
    total = len(rules)

    @debuggable('Seq')
    def _iter(char, parser):
        results = ()

        counter = 0
        for rule in (_build_rule(rule)(char, parser) for rule in rules):
            counter += 1
            while True:
                result, reconsume = next(rule)
                if result is None:
                    yield result, reconsume
                else:
                    break

            results += result

            if counter < total:
                yield None, reconsume

        yield results, reconsume

    return _iter


class __ChrRange(object):
    def __call__(self, begin, end, inverse=False):
        inverse = inverse is True
        rng = xrange(ord(unicode(begin)[0]), ord(unicode(end)[0]) + 1)

        @debuggable('ChrRange')
        def _iter(char, parser):
            if char() is not None and (ord(char()) in rng) is not inverse:
                yield (char(),), False
            raise ParseError(got=char() or '<EOF>', expected=['character in class [{}-{}]'.format(repr(unicode(begin)[0]), repr(unicode(end)[0]))])

        return _iter

    def __getitem__(self, slicee):
        if type(slicee) != slice:
            raise BadRuleException('character range subscripts must be slices of characters')

        return ChrRange(slicee.start, slicee.stop, slicee.step)

ChrRange = __ChrRange()


def Opt(*rules):
    rule = _build_rule(rules)

    @debuggable('Opt')
    def _iter(char, parser):
        grule = rule(char, parser)

        try:
            while True:
                result, reconsume = next(grule)
                if result is not None:
                    yield result, reconsume
                    break

                yield None, reconsume
        except ParseError:
            yield (), True

    return _iter


def Plus(*rules):
    rule = _build_rule(rules)

    @debuggable('Plus')
    def _iter(char, parser):
        results = []

        try:
            while True:
                grule = rule(char, parser)

                while True:
                    result, reconsume = next(grule)
                    if result is not None:
                        results.append(result)
                        break

                    yield None, reconsume

                yield None, reconsume
        except ParseError as e:
            if len(results) == 0:
                raise e  # don't pass the stack; make sure we see that it's from here.

            yield tuple(results), True

    return _iter


def Star(*rules):
    return Opt(Plus(*rules))


def Discard(*rules):
    rule = _build_rule(rules)

    @debuggable('Discard')
    def _iter(char, parser):
        grule = rule(char, parser)

        while True:
            result, reconsume = next(grule)
            if result is not None:
                yield (), reconsume
                break
            yield None, reconsume

    return _iter


def Str(*rules):
    rule = _build_rule(rules)

    @debuggable('Str')
    def _iter(char, parser):
        grule = rule(char, parser)
        while True:
            result, reconsume = next(grule)
            if result is not None:
                result = (''.join(flatten(result)),)
                yield result, reconsume
                break
            yield None, reconsume

    return _iter


@debuggable('Dot')
def Dot(char, parser):
    if char() is None:
        raise ParseError(got='<EOF>', expected=['any non-EOF character'])
    yield (char(),), False


def All(rule, *conditionals):
    if len(conditionals) == 0:
        raise BadRuleException('must supply at least one conditional')

    def _iter(char, parser):
        gconds = [_build_rule(cond)(char, parser) for cond in conditionals]
        grule = _build_rule(rule)(char, parser)

        while True:
            for gcond in gconds:
                reconsume = True
                while reconsume:
                    result, reconsume = next(gcond)
                    if result is not None:
                        raise ParseError(got='conditional result: {}'.format(result), expected=['never returning conditional rule'])

            reconsume = True
            while reconsume:
                result, reconsume = next(grule)
                if result is not None:
                    yield result, reconsume
                    raise StopIteration()

            yield None, False

    return _iter


def In(chars, inverse=False):
    if not chars:
        raise BadRuleException('must supply a string/iterable with at least one character')

    def _iter(char, parser):
        if char() is not None and (char() in chars) is not inverse:
            yield (char(),), False
        raise ParseError(got=char(), expected=['{}one of: {}'.format('not ' if inverse else '', repr(''.join(chars)))])

    return _iter



"""Houses the Pegasus parser class

Unicode characters not being read correctly?
Check out http://stackoverflow.com/a/844443/510036
"""
from __future__ import unicode_literals

import inspect
from itertools import chain as iterchain
from pegasus.rules import _build_rule, ParseError, Lazy


class EmptyRuleException(Exception):
    pass


class NoDefaultRuleException(Exception):
    pass


class NotARuleException(Exception):
    pass


def rule(*rules):
    stack = inspect.stack()[1]
    caller_module = stack[3]
    caller_file = stack[1]

    """Marks a method as a rule"""
    def wrapper(fn):
        _rules = rules
        if len(_rules) == 0:
            raise EmptyRuleException('cannot supply an empty rule')

        Lazy._LOOKUPS[(caller_file, caller_module, fn.__name__)] = fn

        setattr(fn, '_rule', rules)
        return fn

    return wrapper


class Parser(object):
    """The Pegasus Parser base class

    Extend this class and write visitor methods annotated with the @rule decorator,
    create an instance of the parser and call .parse('some str') on it.
    """

    def parse(self, rule, iterable, match=True):
        """Parses and visits an iterable"""
        if not hasattr(rule, '_rule') or not inspect.ismethod(rule):
            raise NotARuleException('the specified `rule\' value is not actually a rule: %r' % (rule,))

        prule = _build_rule(rule)

        itr = iterchain.from_iterable(iterable)
        c = None
        grule = None

        for c in itr:
            reconsume = True
            while reconsume:
                if grule is None:
                    grule = prule(lambda: c, self)

                result, reconsume = next(grule)

                if result is not None:
                    if match:
                        raise ParseError(got='result (rule returned a result without fully exhausting input)')
                    else:
                        return result[0]

        if grule:
            c = None
            reconsume = True
            result = None
            while reconsume:
                result, reconsume = next(grule)
                if result is not None and not match:
                    return result[0]
            return result[0]

        return None



"""A basic functionality test"""
from __future__ import unicode_literals

from pegasus import Parser, rule
from pegasus.rules import Plus, Opt, Discard, Star, ChrRange as C, EOF, Str


class SimpleParser(Parser):
    @rule(['hello', 'greetings', 'salutations', 'yo'])
    def greeting(self, *_):
        pass

    @rule(Str(Plus([C['a':'z'], C['A':'Z']])))
    def name(self, name):
        return name

    @rule(Discard(greeting, Opt(','), Plus(' ')), name, Discard(Star('!')), EOF)
    def hello_world(self, name):
        return name


def test_simple_parser():
    parser = SimpleParser()
    assert 'Paul' == parser.parse(SimpleParser.hello_world, 'hello, Paul!')
    assert 'Sheila' == parser.parse(SimpleParser.hello_world, 'hello,     Sheila')
    assert 'Josh' == parser.parse(SimpleParser.hello_world, 'hello,     Josh!!!')
    assert 'Paul' == parser.parse(SimpleParser.hello_world, 'greetings, Paul!')
    assert 'Sheila' == parser.parse(SimpleParser.hello_world, 'yo,   Sheila!')
    assert 'Josh' == parser.parse(SimpleParser.hello_world, 'salutations,     Josh')



"""A few utilities"""


def flatten(obj, depth=None, current_depth=0):
    if depth is not None and current_depth > depth:
        yield obj
    else:
        for o in obj:
            if type(o) in [tuple, list]:
                for o2 in flatten(o, depth, current_depth + 1):
                    yield o2
            else:
                yield o



from distutils.core import setup
setup(
    name='pegasus',
    packages=['pegasus'],
    version='0.1.2',
    description='Yet another PEG parser generator',
    author='Josh Junon',
    author_email='josh@junon.me',
    url='https://github.com/qix-/pegasus',
    download_url='https://github.com/qix-/pegasus/tarball/0.1.2',
    keywords=['peg', 'packrat', 'pakrat', 'parser', 'generator', 'pegasus', 'visitor', 'expression', 'grammar'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Software Development :: Compilers',
        'Topic :: Software Development :: Interpreters',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Pre-processors',
        'Topic :: Text Editors :: Text Processing',
        'Topic :: Utilities'
    ],
)
